/**
 * A base class for grids that are NOT CPB Components.  Adds the following niceties:
 * 
 * 1. When paging and sorting using a JsonRest store, the grid keeps its size
 *    as opposed to rapidly shrinking/growing while fetching data.
 * 2. If a page selector is visible via the Pagination mixin, it is set to the
 *    proper value for the initial row count (which isn't done by Pagination by default!).
 */
define(
      "sas-cpb/datagrid/BaseGrid", ['../cpb', '../prefs/GridPrefs', '../prefs/SessionPrefs', '../datagrid/MultiColumnSortingMixin',
       'dgrid/Grid', 'dgrid/Selection', 'dgrid/extensions/Pagination', 'dgrid/extensions/ColumnResizer',
       'dgrid/extensions/ColumnReorder', 'dijit/registry',
       'dojo/_base/declare', 'dojo/_base/array', 'dojo/_base/lang', 'dojo/query', 'dojo/aspect', 'dojo/on',
       'dojo/domReady!'],
function(cpb, GridPrefs, SessionPrefs, MultiColumnSortingMixin,
      Grid, Selection, Pagination, ColumnResizer, ColumnReorder, registry,
      declare, array, lang, query, aspect, on) {

   var BaseGrid = declare('BaseGrid', [Grid, Selection, Pagination, MultiColumnSortingMixin, ColumnResizer, ColumnReorder], {

      allowTextSelection: true,
      selectionMode: 'none',
      userInitiatedSort: false,

      constructor: function(args) {

         // Check for cached preferences in session storage, and use them instead of defaults if found.
         // Currently supported: page size, page number.
         if (cpb.preferencesPrefix) {

            this._prefsPrefix = cpb.preferencesPrefix;
            this.pageSizeStorageName = this._prefsPrefix + "_pageSize";
            this.pageNumberStorageName = this._prefsPrefix + "_pageNumber";

            args.rowsPerPage = SessionPrefs.getItem(this.pageSizeStorageName, args.rowsPerPage);
            args.pageNumber = SessionPrefs.getItem(this.pageNumberStorageName, args.pageNumber);
         }

         GridPrefs.loadGridPrefs(args, 1);
         lang.mixin(this, args);

         // Ensure any page size options are sorted, and the initial row count
         // is included in it.
         if (this.pageSizeOptions) {
            if (this.rowsPerPage &&
                  array.indexOf(this.pageSizeOptions, this.rowsPerPage) === -1) {
               this.pageSizeOptions.push(this.rowsPerPage);
            }
            this.pageSizeOptions.sort(function(a, b) { return a - b; });
         }

         // Track whether we are in a user-initiated sort.
         this.userInitiatedSort = false;
      },

      postCreate: function() {

         this.inherited(arguments);
         var grid = this;

         this.on('dgrid-refresh-complete', function(evt) {
            // Let grid resize to whatever's appropriate for its data
            //grid._fixHeight(false);
            if (grid._isSorting) {
               delete grid._isSorting;
            }
            // Resize required to render RHS "scrollbar" area correctly on first render
            grid.resize();
         });

         this.on('dgrid-sort', function(evt) {
            grid.userInitiatedSort = true;
            console.log("userInitiatedSort - begin");
         });

         // Hack to make page size selector default to our initial page size (and have an aria label)
         var pageSizeSelector = this.cpb_getPageSizeSelector();
         if (this.rowsPerPage && pageSizeSelector) {
            pageSizeSelector.value = this.rowsPerPage;
            pageSizeSelector.setAttribute('aria-label', AppProperties['table.rowCount.select.tip']);
         }

         GridPrefs.registerPrefsListener(this);

         //Add hook to store the pageSize if it is changed.
         if (pageSizeSelector) {
            on(pageSizeSelector, 'change', function (evt) {
               if (grid.pageSizeStorageName) {
                  SessionPrefs.setItem(grid.pageSizeStorageName, grid.rowsPerPage);
               }

               // Reset the page number to 1 when the page size changes.  This is what Pagination does anyway, we just
               // need to make sure the session storage value matches.
               if (grid.pageNumberStorageName) {
                  SessionPrefs.setItem(grid.pageNumberStorageName, 1);
               }
            });
         }

         // Cache the total row count from the store
         aspect.after(grid, 'renderArray', function(results) {
            if (typeof results.total === 'number') {
               grid._fixHeight(false);
               grid.resize();
            }
            else {
               results.total.then(function(results) {
                  grid._fixHeight(false);
                  grid.resize();
               });
            }
         }, true);

         // Work around issues in dgrid proper.
         if (document.body.dir === 'rtl') {
            BaseGrid.fixDgridRtlBugs(this);
         }
         BaseGrid.fixA11yBugs(this);

      },

      _destroyChildrenRecursively: function() {
         var widgets = registry.findWidgets(this.domNode);
         array.forEach(widgets, function(widget) {
            widget.destroyRecursive(false);
         });
      },

      _fixHeight: function(fix) {
         // Check to see if we're in the DOM yet.  This could be initial display now that we save page number.
         if (this.domNode) {
            var scrollers = query('.dgrid-scroller', this.domNode);
            if (scrollers && scrollers.length>0) {
               var scroller = scrollers[0];
               var height = fix ? dojo.position(scroller).h + 'px' : '';
               scroller.style.minHeight = height;
            }
         }
      },

      cpb_getPageSizeSelector: function() {
         var selects = query('.dgrid-page-size', this.domNode);
         // Should be 0 or 1
         return selects && selects.length>0 ? selects[0] : null;
      },

      gotoPage: function(page, focusLink) {

         // Another workaround for dgrid issue 752
         if (this.rowsPerPage && this.rowsPerPage.substring) {
            this.rowsPerPage = parseInt(this.rowsPerPage);
         }

         // Don't shrink the grid height while fetching more data from server
         this._fixHeight(true);
         this._destroyChildrenRecursively();
         this.clearSelection(); // Why is this needed?

         var userInitiatedSort = this.userInitiatedSort || false;
         if (this.userInitiatedSort) {
            console.log("userInitiatedSort - end");
            this.userInitiatedSort = false;
         }

         if ((this._sort[0] && page === 1 && userInitiatedSort) || focusLink || page !== 1 || (this._isSorting == undefined && page === 1) ||
               !SessionPrefs.containsKey(this.pageNumberStorageName)) {
            //Setting page on session when we change pages
            if (this.pageNumberStorageName) {
               SessionPrefs.setItem(this.pageNumberStorageName, page);
            }
            return this.inherited(arguments);
         }
         else { //Setting saved page number when creating and not sorting - Pagination refresh always sets page to 1.
             if ((!userInitiatedSort || (this._sort[0] == undefined && page === 1)) && SessionPrefs.containsKey(this.pageNumberStorageName)) {
                 arguments[0] = SessionPrefs.getItem(this.pageNumberStorageName);
             }
             // Always call inherited.
             return this.inherited(arguments);
         }
      },

      _setSort: function(property, descending) {
         if (this._isSorting) { // dgrid issue #635
            return;
         }

         this._isSorting = true;
         // Don't shrink the grid height while fetching more data from server
         this._fixHeight(true);
         this._destroyChildrenRecursively();
         this.inherited(arguments);

      },

      // Extends ColumnReorder._destroyColumns()
      _destroyColumns: function() {
         this._destroyChildrenRecursively();
         this.inherited(arguments);
      }

   });
   
   // summary:
   //   Fixes random RTL issues in dgrid that should really be done
   //   with CSS in the theme.  This is shared so non-BaseGrid dgrids can
   //   consume it also.
   BaseGrid.fixDgridRtlBugs = function(grid) {
      require(['dojo/dom-style', 'dojo/query'], function(domStyle, query) {
         
         // Column selector doesn't move to the left
         var colSelector = query('.dgrid-header-scroll', grid.domNode);
         if (colSelector && colSelector.length>0) {
            domStyle.set(colSelector[0], 'right', 'auto');
            domStyle.set(colSelector[0], 'left', '0');
         }
     
         // Stuff in the footer isn't mirrored
         var statusText = query('.dgrid-status', grid.domNode);
         if (statusText && statusText.length>0) { // Should always === 1
            domStyle.set(statusText[0], 'float', 'right');
         }
         var pageSizeSelector = grid.cpb_getPageSizeSelector();
         if (pageSizeSelector) {
            domStyle.set(pageSizeSelector, 'float', 'left');
         }
         var navLinks = query('.dgrid-navigation', grid.domNode);
         if (navLinks && navLinks.length>0) {
            domStyle.set(navLinks[0], 'float', 'left');
         }
         // Column selector dropdown menu doesn't move to the left
         colSelector = query('.dgrid-hider-menu', grid.domNode);
         if (colSelector && colSelector.length>0) {
            domStyle.set(colSelector[0], 'left', '17px');
            domStyle.set(colSelector[0], 'right', 'auto');
         }
         
      });
   };
   
   // summary:
   //   Fixes random a11y issues in dgrid.  This is shared so non-BaseGrid dgrids can
   //   consume it also.
   BaseGrid.fixA11yBugs = function(grid) {

      var gridDomNode = grid.domNode;

      // Make the "x-y of z results" text non-focusable (dgrid issue #848)
      query('.dgrid-status', gridDomNode).forEach(function(elem) {
         elem.removeAttribute('tabindex');
      });
      
      // Make disabled pagination links non-focusable (also dgrid issue #848)
      aspect.after(grid, '_updateNavigation', function() {
         var disabledNavSpans = query('.dgrid-page-disabled', gridDomNode);
         disabledNavSpans.forEach(function(elem) {
            elem.removeAttribute('tabindex');
         });
         query('span.dgrid-pagination-links span.dgrid-page-link', gridDomNode).forEach(function(link) {
            link.setAttribute('aria-label', 'Go to page ' + link.innerHTML.trim());
         });
      });

      query('button.dgrid-hider-toggle', gridDomNode).forEach(function(button) {
         button.setAttribute('aria-label', 'Grid ' + grid.id + ' column toggle button');
      });

      // Make tab order correct - press column hider "+" button -> tab -> focus cycles through column checkboxes
      // before page size selector dropdown (or whatever comes after grid if it is not there), not after
      var lastChild = gridDomNode.children[gridDomNode.children.length - 1];
      if (lastChild.className === 'dgrid-hider-menu') {
         var nextToLastChild = gridDomNode.children[gridDomNode.children.length - 2];
         if (nextToLastChild.className === 'dgrid-footer') {
            gridDomNode.removeChild(lastChild);
            gridDomNode.insertBefore(lastChild, nextToLastChild);
         }
      }
   };
   
   return BaseGrid;
   
});
