/**
 * A composite relationship component that shows relationships between a Case or Incident
 * (or Party soon?) to Parties.  Adds additional features above and beyond the standard
 * *PartiesGrid components, including:
 * 
 * 1. Tabs to see related parties grouped by relationship type (optional)
 * 2. Ability to add metadata to relationships.
 */
define("sas-ecm/components/EnhancedPartyRelationshipGrid", ['dojo/_base/declare', 'dojo/_base/lang', 'dojo/dom-construct', 'dojo/request/xhr',
        'sas-cpb/cpb', 'sas-cpb/dialog/Modal', 'sas-cpb/datagrid/CPBDataGrid',
         'sas-cpb/datagrid/multigrid/MultiGrid', 'sas-cpb/datagrid/CPBGridUtil',
         'sas-cpb/dialog/YesNoDialog', 'sas-cpb/messages/Messages', 'sas-cpb/util/NLSUtil',
         'dijit/Toolbar', 'dijit/form/Button', 'dijit/form/ToggleButton',
         'dojo/i18n!./nls/EnhancedPartyRelationshipGrid'
         ],
   function(declare, lang, domConstruct, xhr,
         cpb, Modal, CPBDataGrid,
         MultiGrid, CPBGridUtil,
         YesNoDialog, Messages, NLSUtil,
         Toolbar, Button, ToggleButton,
         resources) {
   
var EnhancedPartyRelationshipGrid = declare('EnhancedPartyRelationshipGrid', [MultiGrid], {
   
   constructor: function(args) {
      this._parentEntityType = args.parentEntityType;
      this._dialogScreenId = args.dialogScreenId;
      this._dialogWidth = args.dialogWidth || 800;
      this._dialogHeight = args.dialogHeight || 500;
      this._tableName = args.tableName;
      this._partyKeyField = args.partyKeyField;
      this._solrAvailable = args.params.addExistingClassic !== true;
      this._userActions = args.params.userActions || [];

      this.searchRequiredBeforeCreate = args.params.searchRequiredBeforeCreate || false;
      this._hasSearched = false;
   },
   
   postCreate: function() {
      
      this.inherited(arguments);
      
      // If a new Party was created via "Add New..." button, display a modal to edit its relationship details.
      var ssKey = this._getDetailsForPartySessionStorageKey();
      var partyKey = sessionStorage[ssKey];
      if (partyKey) {
         sessionStorage.removeItem(ssKey);
         this._editRelationshipDetails(partyKey);
      }
      
   },
   
   // summary:
   //   Returns the key in session storage containing a party created for this component's key.  That party
   //   should have its relationship details filled out.
   _getDetailsForPartySessionStorageKey: function() {
      var parentEntityTypeUpper = this._parentEntityType.toUpperCase();
      var parentEntityKey = parentEntityTypeUpper==='CASE' ? Case.KEY : Incident.KEY;
      return this.get('cpbName') + '_relDetailsRequired_' + parentEntityTypeUpper + '_' + parentEntityKey;
   },
   
   createToolbar: function() {
      
      var self = this;
      var toolbar = new Toolbar({}, this.toolbar);
      var gridName = this.get('formattedCpbName');
      var uiContextId = this.get('uiContextId');
      var button;
      
      if (this.getParam('canAddExisting')) {

         if (this.getParam('addExistingClassic')) {
            button = new Button({
               id: self.get('cpbName') + '_addExistingButton',
               iconClass: 'sasAddIcon',
               title: this.getString('add.existing.label.txt'),
               onClick: function () {
                  self._addExistingParty();
               }
            });
            toolbar.addChild(button);
         }

         if (this.getParam('addExistingSolr')) {
            button = new Button({
               id: self.get('cpbName') + '_addExistingViaSolrButton',
               iconClass: 'sasAddIcon',
               title: this.getString('add.existing.label.txt'),
               onClick: function () {
                  self._addExistingParty(false, true);
               }
            });
            toolbar.addChild(button);
         }
      }
      
      if (this.getParam('canAddNew')) {
         button = new Button({
            id: self.get('cpbName') + '_addNewButton',
            iconClass: 'sasNewIcon',
            title: this.getString('add.new.label.txt'),
            onClick: function() {
               self._addNewParty();
            }
         });
         toolbar.addChild(button);
      }
      
      if (this.getParam('canDelete')) {
         button = new Button({
            id: self.get('cpbName') + '_deleteButton',
            iconClass: 'sasDeleteIcon',
            title: this.getString('delete.label.txt'),
            onClick: function() {
               self._deleteSelectedRows();
            },
            sas_selectionButton: true
         });
         toolbar.addChild(button);
      }
      
      if (this.getParam('canToggleDetails')) {
         toolbar.addChild(new ToggleButton({
            id: self.get('cpbName') + '_toggleDetailsButton',
            iconClass: 'sasDeleteColumnIcon',
            title: this.getString('button.toggleDetailsView.tip'),
            checked: true,
            onChange: function(value) {
               this.set('iconClass', value ? 'sasDeleteColumnIcon' : 'sasAddColumnIcon');
               self.set('detailsPaneHidden', !value);
            }
         }));
      }

      if (this._userActions.length > 0) {

         var result = CPBGridUtil.createGridActions(self.get('cpbName'), this._userActions);
         this.gridActions = result.gridActions;
         // TODO
         //var rowActions = result.rowActions;
         //var refreshListeners = result.refreshListeners;

         this.gridActions.forEach(function(action) {
            toolbar.addChild(action);
         });
      }

      return toolbar;
   },
   
   refreshDetailsPanel: function() {
      
      if (!this._activeGrid || !this.detailsPanel) {
         return;
      }
      
      var selection = this._activeGrid.selection;
      var selectedRowCount = Object.getOwnPropertyNames(selection).length;
      
      if (selectedRowCount===0) {
         this.detailsPanel.set('content', this.getNoSelectionMessage());
      }
      else if (selectedRowCount===1) {
         var partyKey = this._getSelectedPartyKey();
         this._showPartyDetailsImpl(partyKey, this.detailsPanel);
      }
      else {
         this.detailsPanel.set('content', ''/*this.getSelectOneRowMessage()*/);
      }
      
   },
   
   _showPartyDetailsImpl: function(partyKey, container, sync) {
      
      var gridName = this.get('formattedCpbName');
      var storeName = this.get('cpbStoreName');
      var uiContextId = this.get('uiContextId');
      var that = this;
      sync = sync || false;
      
      var content;
      xhr('EnhancedPartyRelationship.do', { handleAs: 'text', sync: sync,
            query: { command: 'writeScreenToDiv', dialogScreenId: that._dialogScreenId, partyKey: partyKey,
               gridName: gridName, udfTable: that._tableName, partyKeyField: that._partyKeyField,
               parentUiContextId: uiContextId, storeName: storeName }}).then(function(response) {
         content = that._createDetailsContent(response);
         if (container.domNode) { // dijit, likely ContentPane
            container.set('content', content);
         }
         else { // DOM element
            container.appendChild(content);
         }
      }, function(err) {
         var errorMessage = NLSUtil.resolve('error_loadingData_txt', resources);
         content = that._createDetailsContent(errorMessage);
         if (container.domNode) { // dijit, likely ContentPane
            container.set('content', content);
         }
         else { // DOM element
            container.appendChild(content);
         }
         console.error(err);
      });
      
   },
   
   // summary:
   //   Returns the key of the first selected row's party.  Assumes the caller has ensured that only 1 row
   //   has been selected.
   _getSelectedPartyKey: function() {
      return this._activeGrid.cpb_getSelectedRowValues('PARTY.PARTY_RK')[0];
   },
   
   _createDetailsContent: function(detailsBody) {
      
      // Experimental - mangle id's to prevent collision with original content
      // TODO: Consider having ScreenRenderer not send down ID's and/or cpb.registerWidget() for
      // screens rendered to divs
      var orig = detailsBody;
      detailsBody = detailsBody.replace(/id="([^"]+)"/g, 'id="' + this.get('cpbName') + '_$1"');
      
      var self = this;
      var div = domConstruct.create('div');
      
      if (!this.get('readOnly') && this.getParam('canEdit')) {
         var editToolbar = new Toolbar({});
         editToolbar.addChild(new Button({
            // Always specify a label for accessibility
            label: this.getString('button.editDetails.tip'),
            showLabel: false,
            iconClass: 'sasEditIcon',
            onClick: function() {
               self._editRelationshipDetails();
            }
         }));
         div.appendChild(editToolbar.domNode);
      }
      
      var contentDiv = domConstruct.create('div');
      contentDiv.innerHTML = detailsBody;
      div.appendChild(contentDiv);
      
      return div;
   },
   
   _editRelationshipDetails: function(partyKey) {
      
      var gridName = this.get('formattedCpbName');
      var storeName = this.get('cpbStoreName');
      var uiContextId = this.get('uiContextId');
      partyKey = partyKey || this._getSelectedPartyKey();

      var url = Global.CONTEXT_PATH + '/EnhancedPartyRelationship.do?gridName=' + gridName +
                '&storeName=' + storeName +
                '&dialogScreenId=' + this._dialogScreenId +
                '&partyKey=' + partyKey +
                '&udfTable=' + this._tableName +
                '&partyKeyField=' + this._partyKeyField +
                '&parentUiContextId=' + uiContextId +
                '&windowLayout=' + WindowLayout.MODAL_DIALOG;
      
      var dialogTitleKey = 'dialog_editRelationshipDetails_title_txt';
      if (this.get('readOnly') || !this.getParam('canEdit')) {
         dialogTitleKey = 'dialog_viewRelationshipDetails_title_txt';
         url += '&viewMode=READ_ONLY';
      }
      
      Modal.show({
         url: url,
         width: this._dialogWidth, height: this._dialogHeight,
         title: NLSUtil.resolve(dialogTitleKey, resources)
      });
      
   },
   
   _addExistingParty: function(showNotice, viaSolr) {
      
      this._hasSearched = true;
      
      var gridName = this.get('formattedCpbName');
      var uiContextId = this.get('uiContextId');
      var searchAction = this._getEntityType() + 'PartySearch';
      if (showNotice && !viaSolr) {
         viaSolr = this._solrAvailable===true;
      }

      // Ugh
      if (viaSolr) {
         searchAction += 'ViaSolr';
      }
      else {
         searchAction = 'Enhanced' + searchAction;
      }
      var self = this;
      
      var url = Global.CONTEXT_PATH + '/' + searchAction + '.do?gridName=' + gridName +
                '&uiContextId=' + uiContextId + '&windowLayout=' + WindowLayout.MODAL_DIALOG;
      if (showNotice) {
         url += '&searchBeforeCreate=true';
      }
      Modal.show({
         url: url,
         width: this._dialogWidth, height: this._dialogHeight,
         title: self.getString('dialog.addExisting.title.txt')});
      
   },
   
   _addNewParty: function() {
      
      if (this.searchRequiredBeforeCreate && !this._hasSearched) {
         this._addExistingParty(true);
         return;
      }
      
      var entityType = this._getEntityType();
      var gridName = this.get('formattedCpbName');
      var uiContextId = this.get('uiContextId');
      var sourceKey = window[entityType].KEY;
      
      window[entityType].newParty(uiContextId, sourceKey, false, gridName, true);
   },
   
   _deleteSelectedRows: function() {
      
      var gridName = this.get('formattedCpbName');
      var uiContextId = this.get('uiContextId');
      var self = this;
      
      var dialog = new YesNoDialog({
         message: self.getString('delete.confirmation.txt'),
         onExecute: function() {
            var grid = self.get('activeGrid');
            var keys = grid.cpb_getSelectedRowValues();
            // Send as a single string param and parse on server to avoid ambiguity from multiple values per row
            var relCodes = grid.cpb_getSelectedRowValues('TEMP.PARTY.RELATIONSHIP').join('|');
            App.protectPage();
            dojo.xhrPost({
              url: 'controller/deletePartyRelationships.json',
              content: { grid: gridName, keys: keys, relationshipCodes: relCodes, uiContextId: uiContextId },
              load: lang.hitch(self, self._ajaxSuccess),
              error: lang.hitch(self, self._ajaxError),
              handleAs: 'json'
            });
         },
         onNo: function() { this.hide(); },
         showCancel: false
      });
      
      dialog.show();
      
   },

   _getEntityType: function() {
      return this._parentEntityType.charAt(0).toUpperCase() + this._parentEntityType.substring(1).toLowerCase();
   },
   
   _ajaxSuccess: function(data) {
      this._closeWindow();
      var storeName = this.get('cpbStoreName');
      CPBDataGrid.refreshGridsSharingStore(storeName);
      SaveChangesConfirmation.setPageDirty();
      var cpbName = this.get('cpbName');
      cpb.fieldValueChanged(cpbName);
      App.unprotectPage();
      if (data.errors) {
         Messages.showError(data.errors); // Can take an array
      }
   },

   _ajaxError: function(error) {
      // TODO: Move this to a property
      App.unprotectPage();
      Messages.showError('An error occurred trying to update relationship information: ' + error);
   },

   _closeWindow: function() {
      EnhancedPartyRelationshipGrid.cancel();
   }

});

EnhancedPartyRelationshipGrid.ok = function() {
   
   var form = document.forms[0];
   var params = cpb.getFormParams(form);
   
   form.action = form.action + '?command=commitRelationshipDetailChanges';
   form.submit();
};

EnhancedPartyRelationshipGrid.cancel = function() {
   // Don't load the module twice, reuse the root level one
   window.top.require(['sas-cpb/dialog/Modal'], function(Modal) {
      Modal.hideActiveModal(false);
   });
};

return EnhancedPartyRelationshipGrid;

});