/**
 * @class CommentManager
 * @description Supports creating, editing, and deleting comments. Behaves as the parent wrapper for all helper
 *              widgets. Manages sort order of comments, rendering, and event handling.
 */
define([
    'dojo/mouse',
    'dojo/store/Memory',
    './EditCategoryDialog',
    './ConfirmCancelDialog',
    'dijit/Tooltip',
    'dijit/focus',
    'dojo/ready',
    './NewCategoryDialog',
    './CustomDialog',
    'dijit/registry',
    'dijit/MenuItem',
    'dojo/aspect',
    'dojo/_base/array',
    'dojo/Deferred',
    'dojo/keys',
    'dojo/dom-class',
    '../LoadingSpinner/LoadingSpinner',
    './AttachmentUploader/AttachmentUploader',
    './Comment/Comment',
    './Comment/Attachment',
    './Comment/ReplyForm',
    './CommentData',
    'dijit/_TemplatedMixin',
    'dijit/_WidgetBase',
    'dijit/_WidgetsInTemplateMixin',
    'dijit/layout/ContentPane',
    'dijit/TitlePane',
    'dojo/_base/declare',
    'dojo/_base/lang',
    'dojo/_base/event',
    'dojo/date/locale',
    'dojo/dom-construct',
    'dojo/dom-style',
    'dojo/i18n!./nls/validate',
    'dojo/on',
    'dojo/query',
    'dojo/text!./templates/CommentManager.html',
    'dojo/window',
    'dojox/html/entities',
    'dojox/lang/functional',
    'fraudCommons/util/NLSUtil'
], function (mouse, Memory, EditCategoryDialog, ConfirmCancelDialog, Tooltip, focus, ready, NewCategoryDialog, CustomDialog, registry, MenuItem, aspect, array, Deferred, keys, domClass, LoadingSpinner, AttachmentUploader, Comment, Attachment, ReplyForm, CommentData, TemplatedMixin, WidgetBase, WidgetsInTemplateMixin, ContentPane, TitlePane, declare, lang, event, locale, domConstruct, domStyle, validate, on, query, template, window, entities, functional, NLSUtil) {

    /**
     * @private
     * @function createComment
     * @param {Object} data - Object representing the comment to create.
     *                 Attributes of data include:
     *                          - {string|number} objectId
     *                          - {string|number} objectType
     *                          - {string} [objectName]
     *                          - {number} id
     *                          - {number} parentId
     *                          - {string} subject
     *                          - {string} author
     *                          - {boolean} editable
     *                          - {boolean} deletable
     *                          - {array} attachments
     *                          - {string} text
     *                          - {string|Date} createdOn
     *                          - {function} onComplete
     * @param {boolean} newCategoriesAllowed - True if the user can create new categories.
     * @param {function} onComplete - The callback to execute after server interaction with the comment.
     * @returns {Comment}
     */
    var createComment = function(data, newCategoriesAllowed, readOnly, deletable, editable, categoriesEnabled, onComplete) {
        return new Comment({
            objectId: data.objectId,
            objectType: data.objectType,
            objectName: data.objectName,
            commentId: data.id,
            parentId: data.parentId,
            subject: (data.subject === "__none__") ? NLSUtil.resolve('no_category', validate) : data.subject,
            author: data.author,
            editable: data.editable && editable,
            deletable: data.deletable && deletable,
            categoriesEnabled: categoriesEnabled,
            newCategoriesAllowed: newCategoriesAllowed,
            attachmentsEditable: data.editable,
            attachments: data.attachments,
            text: data.text,
            date: data.createdOn,
            onComplete: onComplete,
            readOnly: readOnly,
            truncationLength: 300
        });
    };

    return declare([WidgetBase, TemplatedMixin, WidgetsInTemplateMixin], {
        templateString: template,
        widgetsInTemplate: true,

        /**
         * @constructs CommentManager
         * @augments dijit/_WidgetBase
         * @augments dijit/_TemplatedMixin
         * @augments dijit/_WidgetsInTemplateMixin
         * @param {string} user - The current user.
         * @param {*} objectType - The object type. Used for retriving, modifying, and storing comments.
         * @param {*} objectId - The ID of the object. Used for retriving, modifying, and storing comments.
         * @param {*} objectName - The name of the object. Used for retriving, modifying, and storing comments.
         * @param {string} [defaultSort="ascending"] - The default sort order for comments on first load.
         * @param {function} [emptyFileCallback] - The callback to execute if an empty file is staged for upload.
         * @param {boolean} [categoriesEnabled=true] - True if categories will be used.
         * @param {boolean} [newCategoriesAllowed=true] - True if the user can add new categories. Will be automatically set to false if categoriesEnabled is false.
         * @param {boolean} [commentsAddable=true] - True if the user can add comments.
         */
        constructor: function (args) {
            this.commentMap = {};
            this.commentHashMap = {};
            this.categoryCountMap = {};
            this.user = args.user;
            this.objectType = args.objectType;
            this.objectId = args.objectId;
            this.objectName = args.objectName;
            this.defaultSort = args.defaultSort;
            this.readOnly = args.readOnly;
            this.emptyFileCallback = args.emptyFileCallback;
            this.defaultCategories = args.defaultCategories || [];
            this.deletable = args.deletable || false;
            this.editable = args.editable || false;
            this.commentsAddable = ('commentsAddable' in args) ? args.commentsAddable : true;
            this.categoriesEnabled = ('categoriesEnabled' in args) ? args.categoriesEnabled : true;
            this.newCategoriesAllowed = (this.categoriesEnabled && ('categoriesEnabled' in args) ? args.newCategoriesAllowed : true);

            this.noCommentsText = NLSUtil.resolve('no_comments', validate);
            this.noSearchResultsText = NLSUtil.resolve('no_search_results', validate);
            this.forbiddenText = NLSUtil.resolve('user_forbidden', validate);
            this.dateSortTitle = NLSUtil.resolve('date_sort_title', validate);

            this.searchLocations = ['text', 'subject', 'author', 'attachments'];

            this.sortAscending = (this.defaultSort && this.defaultSort.toLowerCase() === "descending") ? false : true;
            if (!this.defaultSort) {
                this.defaultSort = this.sortAscending ? "ascending" : "descending";
            }
        },

        postCreate: function () {
            var that = this;
            this.loadComments();

            this.newComment.set('placeholder', NLSUtil.resolve('new_comment_text', validate));
            this.commentSearch.set('placeholder', NLSUtil.resolve('comment_search_placeholder', validate));

            /*
             * If comments are not allowed to be added, do not show the form.
             */
            if (!this.commentsAddable) {
                domStyle.set(this.commentFormContainer.domNode, { 'display': 'none' });
            }

            if(!this.categoriesEnabled) {
                domClass.add(this.domNode, 'categoriesDisabled');
            }

            /*
             * Toggles the sort order of the comments and rearranges them in place.
             */
            var sortByDate = function () {
                that.set('sortAscending', !that.get('sortAscending'));
                array.forEach(that.container.containerNode.childNodes, function (categoryDomNode, key) {
                    var numChildren = categoryDomNode.firstChild.childNodes.length;
                    for (var i = 0; i < Math.floor(numChildren / 2); i++) {
                        var c1 = categoryDomNode.firstChild.childNodes[i];
                        var c2 = categoryDomNode.firstChild.childNodes[numChildren - 1 - i];
                        domConstruct.place(c1, categoryDomNode.firstChild, numChildren - 1 - i);
                        domConstruct.place(c2, categoryDomNode.firstChild, i);
                    }
                });
            };

            /*
             * Returns the indices of the search query in each section of text.
             */
            var findInComment = function (comment, s) {
                var regex = new RegExp(s, "gi"),
                    result,
                    resultsMap = {
                        'text': [],
                        'subject': [],
                        'author': [],
                        'date': [],
                        'attachments': {}
                    };
                array.forEach(that.searchLocations, function (location) {
                    var indices = [];
                    regex.lastIndex = 0;
                    if(location !== 'attachments' && location !== 'subject') {
                        while (result = regex.exec(comment[location])) {
                            indices.push(result.index);
                        }
                        resultsMap[location] = indices;
                    } else if(location === 'attachments') {
                        array.forEach(comment.attachments, function(attachment) {
                            while (result = regex.exec(attachment.name)) {
                                indices.push(result.index);
                            }
                            if(indices.length > 0) {
                                resultsMap.attachments[attachment.id] = indices;
                            }
                        });
                    } else { //location === 'subject'
                        var subject = comment.subject;
                        if(subject === '__none__') {
                            subject = NLSUtil.resolve('no_category', validate);
                        }
                        while (result = regex.exec(subject)) {
                            indices.push(result.index);
                        }
                        resultsMap.subject = indices;
                    }
                });
                return resultsMap;
            };

            /*
             * Hides comments with no search hits
             */
            var performFilter = function (comment, query) {
                var resultsMap = findInComment(comment, query);
                var hitOnAttachments = false;
                functional.forIn(resultsMap.attachments, function(indices, i, _array) {
                    if(Array.isArray(indices) && indices.length > 0) {
                        hitOnAttachments = true;
                    }
                });
                if (query.length === 0 ||
                    hitOnAttachments === true ||
                    resultsMap.text.length > 0 ||
                    resultsMap.subject.length > 0 ||
                    resultsMap.author.length > 0 ||
                    resultsMap.date.length > 0) {
                    that.commentHashMap[comment.id].widget.forEach(function(comment) {
                        comment.set('show', true);
                        comment.set('searchQuery', query);
                        comment.highlight(query);
                    });
                    return true;
                } else {
                    that.commentHashMap[comment.id].widget.forEach(function(comment) {
                        comment.set('show', false);
                        comment.set('searchQuery', query);
                    });
                    return false;
                }
            };

            var performSearch = function(query) {
                var totalCount = 0;
                if (query.length > 0) {
                    array.forEach(that._commentData, function (thread, i, _commentData) {
                        var count = 0;
                        // Pre-HF threads.
                        if (thread && thread.text !== "") {
                            if(performFilter(thread, query)) {
                                count++;
                            }
                            if (thread.replies && thread.replies.length > 0) {
                                array.forEach(thread.replies, function (reply, i, thread) {
                                    if(performFilter(reply, query)) {
                                        count++;
                                    }
                                });
                            }
                        }
                        // Post-HF threads.
                        else if (thread && thread.subject !== "") {
                            if (thread.replies && thread.replies.length > 0) {
                                array.forEach(thread.replies, function (reply, i, thread) {
                                    if(performFilter(reply, query)) {
                                        count++;
                                    }
                                });
                            }
                        }

                        totalCount += count;
                        that.updateCategoryCount(thread.subject, count, true);
                        if(thread.subject !== "__none__") {
                            that.toggleNoResultsInCategoryMessage(thread.subject);
                        }
                    });
                } else {
                    array.forEach(that._commentData, function (thread, i, _commentData) {
                        var count = 0;
                        if (thread.text !== "" && that.commentHashMap[thread.id]) {
                            that.commentHashMap[thread.id].widget.forEach(function(comment) {
                                comment.set('show', true);
                                comment.set("searchQuery", "");
                                comment.unHighlight();
                            });
                            count++;
                        }
                        if (thread.replies && thread.replies.length > 0) {
                            array.forEach(thread.replies, function (reply, i, thread) {
                                if(that.commentHashMap[reply.id]) {
                                    that.commentHashMap[reply.id].widget.forEach(function(comment) {
                                        comment.set('show', true);
                                        comment.set("searchQuery", "");
                                        comment.unHighlight();
                                    });
                                }
                                count++;
                            });
                        }

                        totalCount += count;
                        that.updateCategoryCount(thread.subject, count, false);
                        if(thread.subject !== "__none__") {
                            that.toggleNoResultsInCategoryMessage(thread.subject);
                        }
                    });
                }

                that.updateCategoryCount(NLSUtil.resolve('all_comments_title', validate), totalCount, query.length > 0);

                if(totalCount === 0) {
                    domStyle.set(that.container.domNode, 'display', 'none');
                    domStyle.set(that.noComments.domNode, 'display', 'none');
                    domStyle.set(that.forbidden.domNode, 'display', 'none');
                    domStyle.set(that.noSearchResults.domNode, 'display', 'block');
                } else {
                    domStyle.set(that.container.domNode, 'display', 'block');
                    domStyle.set(that.noComments.domNode, 'display', 'none');
                    domStyle.set(that.forbidden.domNode, 'display', 'none');
                    domStyle.set(that.noSearchResults.domNode, 'display', 'none');
                }
            };

            /*
             * Filter comments when the search query changes
             */
            on(this.commentSearch, 'change', lang.hitch(that, function (query) {
                if(query.length > 0) {
                    /* Special characters should be escaped so that we only filter on the string literal */
                    query = query.replace(/[.*+?^${}()|[\]\\]/g, "\\$&");
                    domClass.add(this.commentSearchContainer, 'clearable');
                    domClass.add(this.iconClear, 'clearable');
                    if (query.trim().length > 0) {
                        domClass.add(this.domNode, 'searchMode');
                    }
                } else {
                    domClass.remove(this.commentSearchContainer, 'clearable');
                    domClass.remove(this.iconClear, 'clearable');
                    domClass.remove(this.domNode, 'searchMode');
                }

                query = query.trim();
                performSearch(query);
            }));

            on(this.commentSearch, 'keyDown', function(e) {
                if(e.keyCode === 13) {
                    e.stopPropagation();
                    e.stopImmediatePropagation();
                    e.preventDefault();
                }
            });

            /*
             * When the X icon is clicked, clear the search box. If
             */
            that.own(on(that.iconClear, 'click', function(e) {
//                that.clearTooltip.hide();
                Tooltip.hide(that.iconClear);
                var query = that.commentSearch.get('value');
                if(query.length > 0) {
                    that.commentSearch.set('value', "");
                }
            }));

            /*
             *  In-place sorting on all comments in their respective categories.
             */
            on(this.dateSort, "click", function (e) {
                lang.hitch(that, sortByDate)();
            });
            on(this.dateSort, 'mousedown', function (e) {
                if (that.commentFormContainer.get('focused')) {
                    lang.hitch(that, sortByDate)();
                }
            });

            /*
             * Adapted from dijit.layout.StackController::onButtonClick()
             *
             * When the new comment form has focus, any mousedown event fired outside of it will cause the pane to lose
             * focus and shrink in size, which makes category tabs hard to click. This hack will change the category tab
             * on mousedown so the user does not need to click twice: once to lose focus, and once to re-click the
             * desired tab.
             */
            on(this.container.tablist, 'mouseDown', function (evt) {
                if (that.commentFormContainer.get('focused')) {
                    var button = registry.getEnclosingWidget(evt.target);
                    if (button !== this.containerNode && !button.disabled && button.page) {
                        for (var target = evt.target; target !== this.containerNode; target = target.parentNode) {
                            if (domClass.contains(target, this.buttonWidgetCloseClass)) {
                                this.onCloseButtonClick(button.page);
                                break;
                            } else if (target === button.domNode) {
                                this.onButtonClick(button.page);
                                break;
                            }
                        }
                    }
                }
            });

            var _toggleSubmitButton = function (text, subject) {
                if (text && text.trim().length > 0 && that.attachmentUploader.submitButton.get('disabled') && subject && subject !== "__newCategory") {
                    that.attachmentUploader.submitButton.set('disabled', false);
                } else if (!text || text.trim().length === 0 || !subject || subject === "__newCategory") {
                    that.attachmentUploader.submitButton.set('disabled', true);
                }
            };

            this.newComment.watch('value', function (name, oldValue, newValue) {
                lang.hitch(that, _toggleSubmitButton)(newValue, that.subject);
            });

            /*
             * Shows the form controls when the text area has focus.
             * Also handles focus styling for the container node, which mimics the active/focus styling of a standard
             * form text input element.
             */
            on(this.commentFormContainer, "focus", function () {
                var validCommentText = that.newComment.get('value').trim().length > 0;
                domStyle.set(that.attachmentUploader.domNode, 'display', 'block');
                domStyle.set(that.cancelButton, 'display', 'block');
                domClass.remove(that.commentFormContainer.domNode, 'emptyTextBox');
            });

            /*
             * Hides the form controls when the text area loses focus.
             * Also handles focus styling for the container node, which mimics the active/focus styling of a standard
             * form text input element.
             */
            on(this.commentFormContainer, "blur", function () {
                var validCommentText = that.newComment.get('value').trim().length > 0;
                if (!validCommentText) {
                    domStyle.set(that.attachmentUploader.domNode, 'display',  'none');
                    domStyle.set(that.cancelButton, 'display', 'none');
                    domClass.add(that.commentFormContainer.domNode, 'emptyTextBox focused');
                }
            });

            /*
             * Sets the category for the comment whenever a category is selected from a dropdown or a category is chosen
             * from the category picker.
             */
            this.attachmentUploader.categorySelect.watch('value', function (name, oldValue, v) {
                that.newCommentSubject.set('value', v);
                that.subject = v;
                lang.hitch(that, _toggleSubmitButton)(that.newComment.get('value'), that.subject);
            });

            /*
             * Clears the comment text when the X button is clicked.
             */
            this.own(on(this.cancelButton, 'click', function (e) {
                var minimize = function () {
                    that.newComment.set('value', "");
                    that.attachmentUploader.uploader.reset();
                    lang.hitch(that, _toggleSubmitButton)(that.newComment.get('value'), that.subject);
                    focus.focus(that.container.domNode);
                    domStyle.set(that.attachmentUploader.uploader.domNode, "display", "inline-block");
                    domStyle.set(that.attachmentUploader.formAttachmentsPane.domNode, {  display: 'none' });
                    domStyle.set(that.attachmentUploader.removeButton.domNode, "display", "none");
                    domStyle.set(that.attachmentUploader.domNode, 'display', 'none');
                    domStyle.set(that.cancelButton, 'display', 'none');
                    domClass.add(that.commentFormContainer.domNode, 'emptyTextBox focused');

                    var options = that.attachmentUploader.categorySelect.get('options');
                    if (that.newCategoriesAllowed && that.attachmentUploader.categorySelect.newestCategory) {
                        var nc = options.pop(); //remove "edit category" temporarily...

                        options.some(function (o, i) { //...and remove new category
                            if (o.value === that.attachmentUploader.categorySelect.newestCategory) {
                                options.splice(i, 1);
                                that.attachmentUploader.categorySelect.newestCategory = undefined;
                                return true;
                            }
                            return false;
                        });

                        nc.label = NLSUtil.resolve('new_category', validate);
                        nc.value = "__newCategory";
                        options.push(nc); //...and push the original "new category" option back onto the end
                    }
                    that.attachmentUploader.categorySelect.set('options', options);
                    that.attachmentUploader.categorySelect.set('value', options[0].value);
                };
                var dialog = new CustomDialog({
                    title: NLSUtil.resolve('confirm_clear_comment_title', validate),
                    body: new ConfirmCancelDialog(),
                    confirmCallback: function () {
                        minimize();
                        dialog.destroy();
                    }
                });

                var options = that.attachmentUploader.categorySelect.get('options');
                if (that.newComment.get('value') || options[options.length - 1].value === "__editCategory" || that.attachmentUploader.uploader.getFileList().length > 0) {
                    dialog.show();
                } else {
                    minimize();
                }
            }));

            /*
             * Sets the value of the category dropdown when a category is chosen from the category picker.
             */
            this.own(aspect.after(this.container, 'selectChild', function (promise, args) {
                if (that.newCategoriesAllowed) {
                    if (args[0] && args[0].__title) {
                        that.attachmentUploader.categorySelect.set('value', args[0].__title || "__newCategory");
                    }
                }
            }));
            this.own(aspect.before(CommentData, 'save', lang.hitch(that, function (node, method) {
                that.currentTab = that.container.get('selectedChildWidget');
            })));
            this.own(aspect.before(CommentData, 'update', lang.hitch(that, function (node, method) {
                that.currentTab = that.container.get('selectedChildWidget');
            })));
            this.own(aspect.before(CommentData, 'deleteComment', lang.hitch(that, function (node, method) {
                that.currentTab = that.container.get('selectedChildWidget');
            })));

            /*
             * Adds focus class styling to the comment box's container node. Since the border around the TextArea in the
             * reply box is not shown, the container should mimic the styling when the TextArea is focused. Note that
             * this styling is only necessary when the TextArea's value is invalid (trimmed text has length 0). Once it
             * contains valid text, the TextArea uses its default styling and this mock-styling is no longer needed.
             */
            this.commentFormContainer.watch('focused', function (name, oldValue, newValue) {
                if (newValue) {
                    domClass.add(that.commentFormContainer.domNode, 'focused');
                } else {
                    domClass.remove(that.commentFormContainer.domNode, 'focused');
                }
            });

            if (this.readOnly) {
                this.set('disabled', true);
            }

            /*
             * Set the styling for the Sort button. Comments sorted by date ascending will show an arrow pointing
             * upwards; those sorted descending will show an arrow pointing downward.
             */
            domClass.add(this.dateSortImg, this.sortAscending ? 'dateSortAscending' : 'dateSortDescending');
            this.watch("sortAscending", function (name, oldValue, newValue) {
                var getSortClass = function (/* Boolean */ b) {
                    return b ? 'dateSortAscending' : 'dateSortDescending';
                };
                domClass.remove(that.dateSortImg, getSortClass(oldValue));
                domClass.add(that.dateSortImg, getSortClass(newValue));
            });

            domClass.add(that.attachmentUploader.categorySelect.dropDown.domNode, 'categorySelectMenu');
        },

        /**
         * @description Retrieves the comments for the current entity and calls render().
         */
        loadComments: function () {
            var that = this,
                completedLoadingFlag = false,
                targetNodeArray = query('.commentContainer'),
                targetNode = registry.byNode(targetNodeArray.length >= 1 ? targetNodeArray[0] : this.domNode);

            this.commentSearch.set('value', "");
            that.clearCommentHashMap();

            /*
             * Spawn a loading spinner and destroy it when the Comment Manager has rendered.
             */
            new LoadingSpinner().target(targetNode).until(function () {
                return completedLoadingFlag;
            }).startSpinners(function () {
            });
            CommentData.retrieve(that.objectId, that.objectType, that.objectName, function (commentData) {
                that._commentData = commentData;
                that.render().then(function () {
                    completedLoadingFlag = true;
                });
            }, function(err) {
                if(err.response && err.response.status && err.response.status === 403) {
                    console.info("FOUR OOOOOH THREE");
                    domStyle.set(that.forbidden.domNode, 'display', 'block');
                    domStyle.set(that.actionToolbar.domNode, 'display', 'none');
                    that.commentSearch.set('disabled', true);
                    that.set('disabled', true);
                    completedLoadingFlag = true;
                }
            });

        },

        /**
         * @description Performs rendering of the widget's components and child widgets.
         * @returns {dojo/Promise} The promise that will resolve when the widget is fully rendered.
         */
        render: function () {
            var that = this;
            var deferred = new Deferred();

            this.showNoCommentsMessage();

            this.container.destroyDescendants();
            that.commentMap = {};
            this.attachmentUploader.uploader.reset();
            domStyle.set(this.attachmentUploader.uploader.domNode, 'display', 'inline-block');

            this.newTopicForm.reset();
            var __loadComments = lang.hitch(this, this.loadComments);

            var hashComment = lang.hitch(this, function (id, comment, widget) {
                if(!that.commentHashMap[id]) {
                    that.commentHashMap[id] = comment;
                    that.commentHashMap[id].widget = [widget];
                } else {
                    that.commentHashMap[id].widget.push(widget);
                }
            });

            functional.forEach(this._commentData, function (category, key) {
                that.commentMap[category.subject] = category;
            });

            this.defaultCategories.forEach(function (/* String */ c) {
                if (!that.commentMap[c]) {
                    var category = {
                        "subject": c,
                        "replies": [],
                        "text": "",
                        "objectType": that.objectType,
                        "objectId": that.objectId,
                        "parentId": 0,
                        "defaultCategory": true
                    };
                    that._commentData.push(category);
                    that.commentMap[c] = category;
                } else {
                    that.commentMap[c].defaultCategory = true;
                }
            });

            /*
             * When no comments exist, display the "no comments" pane; otherwise render the comments.
             */
            if (!this._commentData || this._commentData.length === 0) {
                this.showNoCommentsMessage();
                this.attachmentUploader.categorySelect.addOption({value: "", label: "&nbsp;"});
            } else {
                domStyle.set(this.container.domNode, 'display', 'block');
                domStyle.set(this.noComments.domNode, 'display', 'none');

                var allCommentsContainer = new ContentPane({}),
                    allComments = [],
                    allCommentWidgets = [],
                    categoryList = [];

                /* Sort the category picker alphabetically. "All Comments" will always be on the bottom. */
                this._commentData.sort(function (a, b) {
                    return a.subject.localeCompare(b.subject);
                });

                /*
                 * Loop through each category returned by the server response and generate a new tab for its comments.
                 * If a category has body text, it will be rendered as the first comment. Categories with no body text
                 * only show subsequent replies.
                 */
                this.showComments = false;
                functional.forEach(this._commentData, function (category, key) {
//                    that.commentMap[category.subject] = category;
                    var numComments = category.replies.length;
                    var contentPane = new ContentPane({});

                    var categoryLabel = entities.encode(category.subject);
                    var categoryValue = category.subject;

                    if(category.subject === "__none__") {
                        categoryLabel = NLSUtil.resolve('no_category', validate);
                    }

                    if (key === 0) {
                        that.subject = categoryLabel;
                    }

                    var renderCategory = false;
                    if (category.text.length > 0) {
                        numComments++;
                        allComments.push(category);
                        var c = createComment(category, that.newCategoriesAllowed, that.readOnly, that.deletable, that.editable, that.categoriesEnabled, __loadComments);
                        hashComment(c.get('commentId'), category, c);
                        allCommentWidgets.push(c);
                        contentPane.set('content', c);
                        categoryList.push({label: categoryLabel, value: categoryValue});
                        renderCategory = true;
                    } else if (category.defaultCategory) {
                        categoryList.push({label: categoryLabel, value: categoryValue});
                        renderCategory = true;
                    }

                    if (category.replies && category.replies.length > 0) {
                        allComments = allComments.concat(category.replies);
                        functional.forEach(category.replies, function (reply) {
                            if (reply.text.length > 0) {
                                if (!renderCategory) {
                                    categoryList.push({label: categoryLabel, value: categoryValue});
                                }
                                var c = createComment(reply, that.newCategoriesAllowed, that.readOnly, that.deletable, that.editable, that.categoriesEnabled, __loadComments);
                                hashComment(c.get('commentId'), reply, c);
                                allCommentWidgets.push(c);
                                contentPane.addChild(c, that.sortAscending ? 0 : null);
                                renderCategory = true;
                            }
                        });
                    }

                    that.commentStore = new Memory(Object.keys(that.commentHashMap).map(function (k) {
                        return that.commentHashMap[k];
                    }));

                    if (renderCategory) {
                        /* Anchors category title and comment count to the left and right, respectively */
                        contentPane.set({
                            'title': '<div><span class="commentTabTitle">' + categoryLabel + '</span><span class="commentTabCount">' + numComments + '</span></div>',
                            '__title': categoryValue,
                            'class': 'threadContainer'
                        });
                        that.container.addChild(contentPane);
                        that.categoryCountMap[category.subject] = {
                            count: numComments,
                            searchCount: numComments,
                            contentPane: contentPane
                        };

                        contentPane.noResultsCP = new ContentPane({
                            'content': '<div class="noCategorySearchResultsText">' + NLSUtil.resolve('no_search_results_category', validate) + '</div>',
                            'class': 'noCategorySearchResults'
                        });
                        contentPane.addChild(contentPane.noResultsCP);
                        domStyle.set(contentPane.noResultsCP.domNode, 'display', 'none');

                        on(contentPane, 'show', that.toggleNoResultsInCategoryMessage, category.subject);
                    }

                    that.showComments = that.showComments || renderCategory;
                });
                
                /*
                 * When categories are enabled, populate the dropdown with the available categories
                 */
                if (this.categoriesEnabled) {
                    this.attachmentUploader.categorySelect.set('options', []);
                    if (categoryList.length > 0) {
                        this.attachmentUploader.categorySelect.addOption(categoryList);
                    } else {
                        this.attachmentUploader.categorySelect.addOption({value: "", label: "&nbsp;"});
                    }
                }

                if (that.showComments) {
                    domStyle.set(that.actionToolbar.domNode, 'display', 'block');
                    allCommentWidgets.forEach(function (comment) {
                        comment.set('categories', categoryList);
                        comment.categories = categoryList;
                    });

                    /* Set title with custom layout for category tabs */
                    allCommentsContainer.set({
                        'title': '<div><span class="commentTabTitle">' + NLSUtil.resolve('all_comments_title', validate) + '</span><span class="commentTabCount">' + allComments.length + '</span></div>',
                        '__title': NLSUtil.resolve('all_comments_title', validate),
                        'class': "threadContainer"
                    });

                    that.categoryCountMap[NLSUtil.resolve('all_comments_title', validate)] = {
                        count: allComments.length,
                        searchCount: allComments.length,
                        contentPane: allCommentsContainer
                    };

                    allComments.sort(function (a, b) {
                        var _a = new Date(a.createdOnString);
                        var _b = new Date(b.createdOnString);
                        return _a - _b;
                    });

                    functional.forEach(allComments, function (comment) {
                        var c = createComment(comment, that.newCategoriesAllowed, that.readOnly, that.deletable, that.editable, that.categoriesEnabled, __loadComments);
                        hashComment(c.get('commentId'), comment, c);
                        c.set('categories', categoryList);
                        allCommentsContainer.addChild(c, that.sortAscending ? 0 : null);
                    });

                    this.container.addChild(allCommentsContainer, 'first');
                } else {
                    this.showNoCommentsMessage();
                }
                focus.focus(that.container.domNode);
            }

            /* Set initial values for AttachmentUploader */
            this.attachmentUploader.set({
                d_form: this.newTopicForm,
                editForm: false,
                isNewThread: true,
                commentId: this.commentId,
                emptyFileCallback: this.emptyFileCallback,
                onComplete: function () {
                    that.loadComments();
                },
                dataCallback: function () {
                    return {
                        'objectId': that.objectId,
                        'objectType': that.objectType,
                        'objectName': that.objectName,
                        'subject': that.subject,
                        'parentId': that.commentMap[that.subject] ? that.commentMap[that.subject].id : 0,
                        'text': that.newComment.get('value')
                    };
                }
            });

            this.attachmentUploader.startup();
            this.subject = this.attachmentUploader.categorySelect.get('value');

            if (this.newCategoriesAllowed) {
                this.attachmentUploader.categorySelect.addOption({value: "__newCategory", label: NLSUtil.resolve('new_category', validate)});
            }
            this.container.startup();

            /*entMn
             * Display a tooltip next to the category pane when the user hovers over a tab
             */
            new Tooltip({
                connectId: this.container.tablist.domNode,
                selector: "div.dijitTab",
                getContent: function (node) {
                    //return registry.byNode(node).page.__title;
                    var title = registry.byNode(node).page.__title;
                    if(title === "__none__") {
                        title = NLSUtil.resolve("no_category", validate);
                    }
                    return entities.encode(title);
                }
            });

            this.clearTooltip = new Tooltip({
                connectId: this.commentSearchContainer,
                selector: ".clearable.iconClear",
                label: NLSUtil.resolve("clear_search_tooltip", validate),
                position: ["above"]
            });

            /* Sets the current tab. Mitigates an issue where when the Comment Manager's render() method is called, the
             * tab is reset to whatever the default is when the page loads.
             */
            var children = this.container.getChildren();
            if (children.length > 0) {
                if (this.currentTab) {
                    var index = 0;
                    children.forEach(function (tab, i) {
                        if (tab.__title === that.currentTab.__title) {
                            index = i;
                        }
                    });
                    this.container.selectChild(children[index], false);
                } else {
                    this.container.selectChild(children[0], false);
                }
            }

            return deferred.resolve();
        },

        toggleNoResultsInCategoryMessage: function(title) {
            if (this.categoryCountMap) {
                var categoryObj = this.categoryCountMap[title];

                if (categoryObj && categoryObj.contentPane) {
                    if ((categoryObj.searchCount === 0 || categoryObj.count === 0) && categoryObj.contentPane.noResultsCP) {
                        domStyle.set(categoryObj.contentPane.noResultsCP.domNode, 'display', 'block');
                    } else {
                        domStyle.set(categoryObj.contentPane.noResultsCP.domNode, 'display', 'none');
                    }
                }
            }
        },

        showNoCommentsMessage: function () {
            domStyle.set(this.container.domNode, 'display', 'none');
            domStyle.set(this.actionToolbar.domNode, 'display', 'none');
            domStyle.set(this.noComments.domNode, 'display', 'block');
        },

        hideNoCommentsMessage: function () {
            domStyle.set(this.container.domNode, 'display', 'block');
            domStyle.set(this.actionToolbar.domNode, 'display', 'block');
            domStyle.set(this.noComments.domNode, 'display', 'none');
        },

        updateCategoryCount: function(/* String */ subject, /* Number */ count, /* Boolean */ searchMode) {
            var obj = this.categoryCountMap[subject];
            if(obj) {
                var countStr = searchMode ? NLSUtil.resolve('category_count_search_mode', validate, count, obj.count) : count;
                obj.searchCount = count;
                var mSubject = (subject === '__none__') ? NLSUtil.resolve('no_category', validate) : subject;
                obj.contentPane.set('title', '<div><span class="commentTabTitle">' + mSubject + '</span><span class="commentTabCount">' + countStr + '</span></div>');
            }
        },

        clearCommentHashMap: function() {
            this.commentHashMap = {};
        },

        _setDisabledAttr: function (/* Boolean */ b) {
            domStyle.set(this.commentFormContainer.domNode, 'display', 'none');
            this._set('disabled', true);
        },

        _getDisabledAttr: function () {
            return this.disabled;
        }

    });
});
