/***********************************************************************************/
/* Copyright (c) 2014-2015, SAS Institute Inc., Cary, NC, USA, All Rights Reserved */
/*                                                                                 */
/* NAME:          ecm_srch_create_json                                             */
/* PRODUCT:       Enterprise Case Management                                       */
/* PURPOSE:       create json file for search indexing                             */
/* PARAMETERS:    object_nm=ECM entity type name.                                  */
/*                          CASE, INCIDENT, PARTY, RR, EFILE                       */
/*                type_discriminator_field_name=type_discriminator_field_name      */
/*                          defined in search configuration xml.                   */
/*                unique_id_field=unique_id_field                                  */
/*                          defined in search configuration xml.                   */
/*                type_field_value=type field value defined in search configuration*/
/*                srch_list_dsn= name of table containg the list search fields to  */
/*                               that will be indexed.                             */
/*                data_list_dsn=list of data tables containing data that will be   */
/*                               indexed.                                          */
/*                          <object_nm>_RK must exist in all tables                */
/*                          All tables with data type=LONG must have               */
/*                          UDF_TABLE_NM and UDF_NM.                               */
/*               unprocessed_json_fref=fileref pointing to the unprocessed json    */
/*                                     directory.                                  */
/*               out_list_dsn=Ouput table containing the list of output json files */
/*               output_size_limit=size limit of the output json files.            */
/*               json_prefix=text added as the prefix of the output file names.    */
/*                index_group_YN=Y to add '__group' to the json file.              */
/* NOTE:                                                                           */
/*               The output file names will be <json_prefxi>_<file number>.json    */
/*               The output file may be slightly larger than the limit because     */
/*               an entire object has to be output before starting a new output    */
/*               file.                                                             */
/***********************************************************************************/
 
%macro ecm_srch_create_json(   object_nm=
                             , type_discriminator_field_name=
                             , unique_id_field=
                             , type_field_value=
                             , srch_list_dsn=
                             , data_list_dsn=
                             , multi_row_yn=
                             , unprocessed_json_fref=
                             , out_list_dsn=json_list
                             , output_size_limit=
                             , json_prefix=
                             , groups="Ent Case Mgmt Users"
                             , index_group_YN=Y
                            );


%let abortRC=0;

/* Output directory path is required */
%let unprocessed_json_dir=;
%if %str(&unprocessed_json_fref)=%str() %then
%do;
     %let abortRC=1;
     %put %sysfunc(sasmsg(&smdds,GEN_PARM_MISSING_ERROR,NOQUOTE,UNPROCESSED_JSON_FREF)) ;
     %goto mEXIT;
%end; 
%let unprocessed_json_dir=%sysfunc(pathname(&unprocessed_json_fref,F));
%if %nrbquote(&unprocessed_json_dir) = %nrbquote() %then 
%do;
     %let abortRC=1;
     %put %sysfunc(sasmsg(&smdds,GEN_PARM_INVALID_ERROR,NOQUOTE,UNPROCESSED_JSON_FREF,&unprocessed_json_fref)) ;
     %goto mEXIT;
%end; 

/* call ecm_mkdir to either create the directory or clear the existing directory */

 %local createdRC;
 %let createdRC=0;
    %ecm_mkdir(pathname=%superq(unprocessed_json_dir),rcvar=createdRC,resetYN=Y);

    %if &createdRC ne 1 %then 
    %do; 
        %let successYN=N;
        %put %sysfunc(sasmsg(&smdds,GEN_MKDIR_ERROR,NOQUOTE,%superq(outpath))) ; 
        %goto mExit; 
    %end;

/* create a new temp file if it does not exist */
%if not %sysfunc(fexist(tmp1)) %then 
%do;
    filename tmp1 temp;
%end;

/* generate sas statements for writing out data fields in a rectangular structured table */
%let rect_table_nm=;
%if %upcase(&multi_row_yn) ne Y %then
%do;


data _null_;
    retain mydoublequote '@';
    file tmp1;    
    set &srch_list_dsn end=eof;
    where table_struct_type='RECT' ;
    length sas_stmt $ 32760;
    
    sas_stmt=', "'||strip(srch_field_nm)||'" :';
    put "json_stmt='" sas_stmt "'; link WRT_JSON;";
    sas_stmt='json_stmt=';
    put sas_stmt;
    if sas_datatype=1 and srch_datatype ne 'DATE' then
    do; 
       sas_stmt='ifc('  
                || strip(src_column_nm)
                || '=.,"null",strip(put('
                ||strip(src_column_nm)
                ||',best.)));';
    end;
    else if sas_datatype=1 and srch_datatype='DATE' then
    do; 
         sas_stmt='ifc('  
                || strip(src_column_nm)
                || '=.,"null",''"''||strip(put('
                ||strip(src_column_nm)
                ||',E8601LX25.))||''"'');';
          
    end;
    else if srch_datatype='BOOLEAN' then 
    do;
          sas_stmt='ifc('  
                   || strip(src_column_nm)
                   || '="","null",ifc('
                   ||strip(src_column_nm)
                   ||'="1","true","false"));'
                   ;
    end;
    else 
    do;

          sas_stmt='ifc('  
                   || strip(src_column_nm)
                   || '="","null",''"''||strip(tranwrd(tranwrd(tranwrd(compress('
                   ||strip(src_column_nm)
                   ||",'','kw'),'\','\\'),'/','\/'),'@','\@')"
                   ||")||'@');";
          sas_stmt=tranwrd(sas_stmt,mydoublequote,'"');

            
         
    end;
    put sas_stmt;

    sas_stmt='link WRT_JSON;';
    put sas_stmt;

    if eof then call symput('RECT_TABLE_NM',strip(src_table_nm));
run;



%if &syserr > 4 %then 
%do;
     %let abortRC=1;
     %put %sysfunc(sasmsg(&smdds,GEN_MAC_ABORT,NOQUOTE,&sysmacroname)) ;
     %goto MEXIT;
%end;

%end;


%let rect_dsn=;
%let group_dsn=;
%let orig_src_dsn_list=;
%let multi_dsn=;
%let unique_id_field_exist=N;

%if %nrbquote(&json_prefix)=%nrbquote() %then %let json_prefix=J_;



proc sql noprint;
      

      /* get the user group table name */
      select src_dsn into :group_dsn
      from &data_list_dsn 
      where  src_type='GROUP'
      ;

      
      /* only one Rectangular table is supported */
      select distinct src_dsn into :rect_dsn
      from &data_list_dsn 
      where  src_type='RECT'
      ;

      /* Get the multi row table which is properly transposed */
      select distinct src_dsn into :multi_dsn
      from &data_list_dsn 
      where  src_type='MULTI'
      ;
      
      /* Pull the list of UDF value tables */
      select distinct  src_dsn                 
           into
               :long_dsn_1 - 
      from &data_list_dsn
      where  src_type='LONG'
      ;

      /* get the list of names (CHAR, LGCHAR, DATE, NUM) of UDF value tables */
      select distinct  orig_src_dsn                 
           into
               :orig_src_dsn_list separated by ' ' 
      from &data_list_dsn
      where  src_type in ('LONG' )
      ;

      %let long_dsn_cnt=&sqlobs;

      select distinct 'Y' into :unique_id_field_exist 
      from &udf_list_dsn
      where srch_field_nm="&unique_id_field"
;



%if &syserr > 4 %then 
%do;
     %let abortRC=1;
     %put %sysfunc(sasmsg(&smdds,GEN_MAC_ABORT,NOQUOTE,&sysmacroname)) ;
     %goto MEXIT;
%end;
 

%local i_dsn i_nm nm_cnt;

%let nm_cnt=0;
/* get the search field names under UDF value tables */

proc sql noprint;
select distinct srch_field_nm 
       into  :srch_field_nm_1 - 
        from &srch_list_dsn
        where table_struct_type in ('LONG' 'MULTI')
        ;
        %let nm_cnt=&sqlobs;
quit;

%if &syserr > 4 %then 
%do;
     %let abortRC=1;
     %put %sysfunc(sasmsg(&smdds,GEN_MAC_ABORT,NOQUOTE,&sysmacroname)) ;
     %goto MEXIT;
%end;

%if &object_nm=NONE %then %let by_key=%scan(&unique_id_field,2,.);
%else  %let by_key=&object_nm._rk;
   
data  &out_list_dsn(keep=path_nm file_nm); 
/*Ccreate an array of the search field names which values can come from UDF value tables */
/* The arrays are defined to remember which search fields have no UDF values.            */
/* Then, null fields will be generated.                                                  */
/* nm_cnt > 0 is checked because there may not be any UDF value tables (LONG).           */
%if &nm_cnt > 0 %then
%do;
    array srch_fld_nm{&nm_cnt} $ 1024 _temporary_  
        (
    %do i_nm=1 %to &nm_cnt;
        "&&&srch_field_nm_&i_nm"
    %end;
        ) ;
    array src_col_flg{&nm_cnt}  _temporary_ ;
%end;

    length json_stmt $ 32760;
    retain json_stmt;
    retain output_size 0;
    retain file_cnt 0
    retain fid   0;
    retain new_header_flg 0;
    length file_prefix
           file_nm
           path_nm 
           $1024;
    retain file_prefix;

    %if %str(&rect_dsn) = %str() %then
      %do;  /* rectangular table always come first because there is only one record per object*/ 
             in_rect=0;
      %end; 


    set 
      %if %str(&rect_dsn) ne %str() %then
      %do;  /* rectangular table always come first because there is only one record per object*/ 
             &rect_dsn (in=in_rect)
      %end; /* rectangular table always come first because there is only one record per object*/
      
      %if %str(&group_dsn) ne %str() %then
      %do;
        &group_dsn (in=in_group)
      %end;
     
        
      %if %str(&multi_dsn) ne %str() %then
      %do;
        &multi_dsn
      %end;
        %do i_dsn=1 %to &long_dsn_cnt;
            &&&long_dsn_&i_dsn  
        %end;
        end=eof;
    by &by_key
        udf_table_nm 
        udf_nm
     %if %str(&group_dsn)=%str() and &long_dsn_cnt=0 %then
     %do;  /* the source table does not need to presorted if it does not involve multiple tables join */
        notsorted
     %end;
    ;
    if _n_=1 then 
    do; /* first input record  */
        file_prefix="%nrbquote(&json_prefix)";
        link HEADER;
    end; /* first input record  */

    /* first input record of the object */
    if  first.&by_key then link OBJECT_HEADER;

    
    if first.udf_nm and not in_rect then link UDF_HEADER;
    
    link UDF_DETAIL;

    if last.udf_nm then link UDF_TRAILER;
 

    if last.&by_key then link OBJECT_TRAILER;


    if eof then link TRAILER;
    return;


/* create new output file */

 NEW_FILE:
    if fid > 0 then link TERM_FILE;
    file_cnt+1;
    file_nm=strip(file_prefix)||strip(put(file_cnt,best.))||'.json';
    rc=filename('OUTFILE',"%nrbquote(&UNPROCESSED_JSON_DIR)"||"&slash"||strip(file_nm));
    fid=fopen('OUTFILE','O',,'V');
    if fid=0 then 
    do;
       call symput('abortRC','1');
       abort;
    end;
    path_nm="%nrbquote(&UNPROCESSED_JSON_DIR)"||"&slash";
    output;
    output_size=0;
 return;

 /* Close output file */

 TERM_FILE:
    if fid > 0 then fid=fclose(fid);
    rc=filename('OUTFILE');
 return;

/* Add open Json doc tag */
 HEADER:
    link NEW_FILE;
    json_stmt='[';
    link WRT_JSON;
    new_header_flg=1;
 return;

 /* Add close Json doc tag */
 TRAILER:
    if fid > 0 then 
    do;
    json_stmt=']';
    link WRT_JSON;
    link TERM_FILE;
    end;
return; 

/* Add open object tag.                                                                       */
/* Check output size.                                                                         */
/* If it is too big, add close Json doc tag to the existing file, then open a new output file */
/* and add open Json doc tag to the beginning of the output file.                             */

OBJECT_HEADER:
    /* clear flg for checking whether the data field has values */
%if &nm_cnt > 0 %then 
%do;
    do i=1 to dim(src_col_flg);
        src_col_flg{i}=0;
    end;
%end;
    if output_size > &output_size_limit then 
    do;
       link TRAILER;
       link HEADER;
    end;
    if new_header_flg then
    do; 
       json_stmt='{';
       new_header_flg=0;
    end;
    else
    do;
       json_stmt=',{';
    end;
    link WRT_JSON;   
    json_stmt='"'||strip("%nrbquote(%trim(&type_discriminator_field_name))")||'" : "'||"%nrbquote(%trim(&type_field_value))"||'"';
    link WRT_JSON;
%if &unique_id_field_exist=N %then
%do;     
    %if &object_nm=NONE %then
    %do;
    json_stmt=',"'||strip("%nrbquote(&unique_id_field)")||'" : '||strip(&by_key);
    %end;
    %else
    %do;
    json_stmt=',"'||strip("%nrbquote(&unique_id_field)")||'" : '||strip(put(&by_key,best.));
    %end;

    link WRT_JSON;
%end;
%if &index_group_YN=Y and %nrbquote(&groups) ne %nrbquote() and &object_nm=NONE %then
%do;     
    json_stmt=',"groups":'||strip(symgetc('groups'));
    link WRT_JSON;
%end;

    %if %str(&rect_table_nm) ne %str() %then %inc tmp1;;
     
return;
     
/* Add null values for the fields that are in the search list but not in the UFG value tables &/
/* Add close Object tag */
OBJECT_TRAILER:
%if &nm_cnt > 0 %then
%do;
    do i=1 to dim(src_col_flg);
        if src_col_flg{i}=0 then 
        do;
            json_stmt=', "'||strip(srch_fld_nm{i})||'" : null';
            link WRT_JSON;
        end;
    end;
%end;
    json_stmt='}';
    link WRT_JSON;

return;

/* Add search field name */
/* Add open array tag if there are multiple values */
UDF_HEADER:
    if in_group then json_stmt=', "groups":';
    /*
    else json_stmt=', "'||strip(upcase(udf_table_nm))||'.'||strip(upcase(udf_nm))||'" : ';
    */
     else json_stmt=', "'||strip(srch_field_nm)||'" : ';
    link WRT_JSON; 


    if not last.udf_nm then json_stmt=strip(json_stmt)||'[';
    if not in_rect then link WRT_JSON_CONT;

return;

/* Close the array tag if there are multiple values */

UDF_TRAILER:

    if not first.udf_nm then json_stmt=']';

    if not in_rect then link WRT_JSON;

return;

/* Add the UDF or group values to JSON statement */
UDF_DETAIL:
    
    
  %if &nm_cnt > 0 %then
  %do;
  
    i_col=whichc(srch_field_nm, of srch_fld_nm[*]);
    if i_col > 0 then src_col_flg{i_col}=1; 
  %end;

    if not first.udf_nm then comma=',';
    else comma='' ;
    if in_group then 
        json_stmt=strip(json_stmt)||comma||' "'||strip(user_group_nm)||'" ';
%if %index(&orig_src_dsn_list,LGCHR) %then
%do;
    else if ecm_datatype in ('LNGVARCHAR')  then 
       json_stmt=strip(json_stmt)
                ||comma
                ||' "'||tranwrd(tranwrd(tranwrd(strip(compress(udf_value_lgchr,'','kw')),'\','\\'),'"','\"'),'/','\/')
                 ||'" ';
%end;
%if %index(&orig_src_dsn_list,DATE) %then
%do;

    else if ecm_datatype in ('TIMESTAMP' 'DATE')  then
             json_stmt=strip(json_stmt)||comma||' "'||put(udf_value_date,E8601LX25.)||'" ';
%end;
%if %index(&orig_src_dsn_list,NUM) %then
%do;
    else if ecm_datatype in ('INTEGER' 'DOUBLE' 'BIGINIT') then
              json_stmt=strip(json_stmt)||comma||' '||put(udf_value_num,best.)||' ';
%end;
%if %index(&orig_src_dsn_list,CHAR) %then
%do;

else json_stmt=strip(json_stmt)
               ||comma
               ||' "'
               ||tranwrd(tranwrd(tranwrd(strip(compress(udf_value_char,,'kw')),'\','\\'),'"','\"'),'/','\/')
               ||'" '; 
%end;

%if %str(&orig_src_dsn_list)=%str() %then
%do;
     if sas_datatype > . then
     do;
         if sas_datatype=2 then 
            json_stmt=strip(json_stmt)
                      ||comma
                      ||' "'
                      ||tranwrd(tranwrd(tranwrd(strip(compress(udf_value_char,,'kw')),'\','\\'),'"','\"'),'/','\/')
                      ||'" '; 
         else if upcase(srch_datatype) = 'DATE' then  
                 json_stmt=strip(json_stmt)||comma||' "'||put(udf_value_num,E8601LX25.)||'" ';
               else
                 json_stmt=strip(json_stmt)||comma||' '||put(udf_value_num,best.)||' ';
     end;
%end;

 

    if not in_rect then 
    do;
       if not last.udf_nm then link WRT_JSON;
       else link WRT_JSON_CONT;
    end;
     
return;

/* write the json statement and accumulate the output size */
WRT_JSON:
       output_size+length(json_stmt);
       rc=fput(fid,strip(json_stmt));
       rc=fwrite(fid);
       json_stmt='';
return;

/* pushing the json statement without writing it, i.e. no line feed */
WRT_JSON_CONT:
       output_size+length(strip(json_stmt));
       rc=fput(fid,strip(json_stmt));
       json_stmt='';
return;

run;

%if &syserr > 4 %then 
%do;
     %let abortRC=1;
     %put %sysfunc(sasmsg(&smdds,GEN_MAC_ABORT,NOQUOTE,&sysmacroname)) ;
     %goto MEXIT;
%end;

%if &ecm_debug ne Y %then 
%do;
    filename tmp1 clear;
%end;


%mEXIT:

%put %sysfunc(sasmsg(&smdds,GEN_MAC_END_NOTE,NOQUOTE,&sysmacroname)) ;
%mend;

 
